#########################################################################
# Source: Who should review my code? A file location-based code-reviewer recommendation approach for modern code review, SNAER 2015
# Link: https://github.com/sulunemre/rstrace-replication/blob/master/src/Revfinder.py
# Descriptions: String comparison techniques for file path similarity
# Input: The arguments f1, f2 are strings of file path
# Output: Number of common file path components in f1 and f2
# Written By: Patanamon Thongtanunam (patanamon-t@is.naist.jp)
#########################################################################
import pandas as pd
import concurrent.futures

def path2List(fileString):
    '''
    :param fileString: file path string (string)
    :return: a list of file path nodes (list)
    '''
    return fileString.split("/")

'''Four file path string comparison methods'''
def LCP(f1, f2):
    f1 = path2List(f1)
    f2 = path2List(f2)
    common_path = 0
    min_length = min(len(f1), len(f2))
    for i in range(min_length):
        if f1[i] == f2[i]:
            common_path += 1
        else:
            break
    return common_path


def LCSuff(f1, f2):
    f1 = path2List(f1)
    f2 = path2List(f2)
    common_path = 0
    r = range(min(len(f1), len(f2)))
    r = reversed(r)
    for i in r:
        if f1[i] == f2[i]:
            common_path += 1
        else:
            break
    return common_path


def LCSubstr(f1, f2):
    f1 = path2List(f1)
    f2 = path2List(f2)
    common_path = 0
    if len(set(f1) & set(f2)) > 0:
        mat = [[0 for x in range(len(f2)+1)] for x in range(len(f1)+1)]
        for i in range(len(f1)+1):
            for j in range(len(f2)+1):
                if i == 0 or j == 0:
                    mat[i][j] = 0
                elif f1[i-1] == f2[j-1]:
                    mat[i][j] = mat[i-1][j-1] + 1
                    common_path = max(common_path, mat[i][j])
                else:
                    mat[i][j] = 0
    return common_path


def LCSubseq(f1, f2):
    f1 = path2List(f1)
    f2 = path2List(f2)
    if len(set(f1) & set(f2)) > 0:
        L = [[0 for x in range(len(f2)+1)] for x in range(len(f1)+1)]
        for i in range(len(f1)+1):
            for j in range(len(f2)+1):
                if i == 0 or j == 0:
                    L[i][j] = 0
                elif f1[i-1] == f2[j-1]:
                    L[i][j] = L[i-1][j-1] + 1
                else:
                    L[i][j] = max(L[i-1][j], L[i][j-1])
        common_path = L[len(f1)][len(f2)]
    else:
        common_path = 0
    return common_path

'''RevFinder approach'''
def FilePathSimilarity(newFile, pastFile, method):
    '''
    :param newFile: (string)
    :param pastFile:(string)
    :param method: don't need to self-define this parameter
    :return: (int)
    '''
    score = -1
    if method == 'LCP':
        score = LCP(newFile, pastFile)
    elif method == 'LCSuff':
        score = LCSuff(newFile, pastFile)
    elif method == 'LCSubstr':
        score = LCSubstr(newFile, pastFile)
    elif method == 'LCSubseq':
        score = LCSubseq(newFile, pastFile)
    else:
        raise ValueError(
            f"{method} is not valid. Method can be: LCP, LCSuff, LCSubstr, LCSubseq")
    return score

def recommendReviewersByMethod(pastCommit_paths, newCommit_paths, pastReviewers, method):
    recommendedReviewers = {}  # dicts {reviewerName: score}
    temp = newCommit_paths
    newCommit_paths = []
    newCommit_paths.append(temp)
    # Compute review similarity score between newCommit and pastCommit
    score = 0
    for newFile in newCommit_paths:
        for pastFile in pastCommit_paths:
            score += FilePathSimilarity(newFile, pastFile, method)
    score = score / (len(newCommit_paths) * len(pastCommit_paths))

    # Propagate review similarity scores to code-reviewers who involved in a previous review Rp
    if score > 0:
        for reviewer in pastReviewers:
            if reviewer in recommendedReviewers:
                recommendedReviewers[reviewer] += score
            else:
                recommendedReviewers[reviewer] = score

    # Sort by score and return only names
    if len(recommendedReviewers) != 0:
        {k: v for k, v in sorted(recommendedReviewers.items(), key=lambda item: item[1], reverse=True)}

    '''[Another way to implement RevFinder]
    for reviewer in pastReviewers:
        newCommitFiles = newCommits
        pastCommitFiles = pastCommits
        score = 0
        for newFile in newCommitFiles:
            for pastFile in pastCommitFiles:
                score += FilePathSimilarity(newFile, pastFile, method)
        score = score / (len(newCommitFiles) * len(pastCommitFiles))
        if reviewer in recommendedReviewers:
            recommendedReviewers[reviewer] += score
        else:
            recommendedReviewers[reviewer] = score

    # Sort by score and return only names
    if len(recommendedReviewers) != 0:
        {k: v for k, v in sorted(recommendedReviewers.items(), key=lambda item: item[1], reverse=True)}
    '''
    return [*recommendedReviewers]

def recommendReviewers(pastCommits, newCommits, pastReviewers):
    """Runs four different methods in parallel and combines their results
    Arguments:pastCommits {[type]} -- [description]
        newCommit {[type]} -- [description];
        k -- Top K reviewers
    Returns: [type] -- [description]
    """
    # Run in different processes
    with concurrent.futures.ProcessPoolExecutor(max_workers=4) as executor:
        future1 = executor.submit(recommendReviewersByMethod, pastCommits, newCommits, pastReviewers, 'LCP')
        future2 = executor.submit(recommendReviewersByMethod, pastCommits, newCommits, pastReviewers, 'LCSuff')
        future3 = executor.submit(recommendReviewersByMethod, pastCommits, newCommits, pastReviewers, 'LCSubstr')
        future4 = executor.submit(recommendReviewersByMethod, pastCommits, newCommits, pastReviewers, 'LCSubseq')

        l1 = future1.result()
        l2 = future2.result()
        l3 = future3.result()
        l4 = future4.result()

    # Combination
    reviewerScores = {}

    for i, reviewer in enumerate(l1):
        score = len(l1) - i
        if reviewer in reviewerScores:
            reviewerScores[reviewer] += score
        else:
            reviewerScores[reviewer] = score

    for i, reviewer in enumerate(l2):
        score = len(l2) - i
        if reviewer in reviewerScores:
            reviewerScores[reviewer] += score
        else:
            reviewerScores[reviewer] = score

    for i, reviewer in enumerate(l3):
        score = len(l3) - i
        if reviewer in reviewerScores:
            reviewerScores[reviewer] += score
        else:
            reviewerScores[reviewer] = score

    for i, reviewer in enumerate(l4):
        score = len(l4) - i
        if reviewer in reviewerScores:
            reviewerScores[reviewer] += score
        else:
            reviewerScores[reviewer] = score

    # Sort by score and return only names
    if len(reviewerScores) != 0:
        {k: v for k, v in sorted(reviewerScores.items(), key=lambda item: item[1], reverse=True)}
    final_list = [*reviewerScores]

    return final_list